local M = {}

-- stylua: ignore start
M.setup = function()
  -- better navigation for wrapped lines
  vim.keymap.set("n", "j", "gj")
  vim.keymap.set("n", "k", "gk")

  -- maintain cursor position when joining lines
  vim.keymap.set("n", "J", "mzJ`z")

  -- retain selection when making changes in visual mode
  vim.keymap.set("v",  "<c-a>",  "<c-a>gv")
  vim.keymap.set("v",  "<c-x>",  "<c-x>gv")
  vim.keymap.set("v", "g<c-a>", "g<c-a>gv")
  vim.keymap.set("v", "g<c-x>", "g<c-x>gv")
  vim.keymap.set("v",      ">",  "><cr>gv")
  vim.keymap.set("v",      "<",  "<<cr>gv")

  -- place destination of important movements in the center of the screen
  vim.keymap.set("n",     "n",     "nzzzv")
  vim.keymap.set("n",     "N",     "Nzzzv")
  vim.keymap.set("n",     "*",     "*zzzv")
  vim.keymap.set("n",     "#",     "#zzzv")
  vim.keymap.set("n",    "g*",    "g*zzzv")
  vim.keymap.set("n",    "g#",    "g#zzzv")
  vim.keymap.set("n", "<c-d>", "<c-d>zzzv")
  vim.keymap.set("n", "<c-u>", "<c-u>zzzv")

  -- easier window navigation
  vim.keymap.set("n", "<c-j>", "<c-w>j")
  vim.keymap.set("n", "<c-k>", "<c-w>k")
  vim.keymap.set("n", "<c-h>", "<c-w>h")
  vim.keymap.set("n", "<c-l>", "<c-w>l")

  -- disable highlight until next search
  vim.keymap.set("n", "<leader>h", "<cmd>nohlsearch<cr>")

  local window = require("fschauen.util.window")

  -- window resizing
  vim.keymap.set("n",    "<s-Up>",    window.resize_up(2), { desc = "Resize window upward"    })
  vim.keymap.set("n",  "<s-Down>",  window.resize_down(2), { desc = "Resize window downward"  })
  vim.keymap.set("n",  "<s-Left>",  window.resize_left(2), { desc = "Resize window leftward"  })
  vim.keymap.set("n", "<s-Right>", window.resize_right(2), { desc = "Resize window rightward" })

  -- easy tab navigation
  vim.keymap.set("n", "<Right>", "<cmd>tabnext<cr>")
  vim.keymap.set("n",  "<Left>", "<cmd>tabprevious<cr>")

  -- move lines up and down
  vim.keymap.set("n", "<c-a-j>", [[:move .+1<cr>==]])
  vim.keymap.set("n", "<c-a-k>", [[:move .-2<cr>==]])
  vim.keymap.set("v", "<c-a-j>", [[:move '>+1<cr>gv=gv]])
  vim.keymap.set("v", "<c-a-k>", [[:move '<-2<cr>gv=gv]])
  vim.keymap.set("i", "<c-a-j>", [[<esc>:move .+1<cr>==gi]])
  vim.keymap.set("i", "<c-a-k>", [[<esc>:move .-2<cr>==gi]])

  -- move to begin/end of line in insert mode
  vim.keymap.set("i", "<c-a>", "<c-o>^")
  vim.keymap.set("i", "<c-e>", "<c-o>$")

  -- move to begin of line in command mode (<c-e> moves to end by default)
  vim.keymap.set("c", "<c-a>", "<c-b>")

  -- more convenient way of entering normal mode from terminal mode
  vim.keymap.set("t", [[<c-\><c-\>]], [[<c-\><c-n>]])

  -- recall older/recent command-line from history
  vim.keymap.set("c", "<c-j>", "<down>")
  vim.keymap.set("c", "<c-k>", "<up>")

  -- trigger InsertLeave when leaving Insert mode with ctrl-c (see :help i_CTRL-C)
  vim.keymap.set("i", "<c-c>", "<esc>")

  -- quickly change background
  vim.keymap.set("n", "<leader>bg", [[<cmd>let &background = &background ==? "light" ? "dark" : "light"<cr>]])

  -- don't loose the original yanked contents when pasting in visual mode
  vim.keymap.set("x", "<leader>p", [["_dP]])

  local diagnostic = require("fschauen.diagnostic")
  local ui = require("fschauen.util.icons").ui

  -- navigate diagnostics
  vim.keymap.set("n", "<leader>dd", diagnostic.toggle,       { desc = ui.Diagnostic.." [d]iagnostic enable/[d]isable" })
  vim.keymap.set("n", "<leader>do", diagnostic.open_float,   { desc = ui.Diagnostic.." [d]iagnostic [o]pen" })
  vim.keymap.set("n", "<leader>dh", diagnostic.hide,         { desc = ui.Diagnostic.." [d]iagnostic [h]ide" })
  vim.keymap.set("n", "<leader>ds", diagnostic.set_severity, { desc = ui.Diagnostic.." [d]iagnostic [s]everity" })

  -- toggle quickfix and loclist
  vim.keymap.set("n",      "<leader>q", window.toggle_quickfix, { desc = ui.Toggle.." toggle quickfix" })
  vim.keymap.set("n", "<localleader>q", window.toggle_loclist,  { desc = ui.Toggle.." toggle loclist"  })

  local options = require("fschauen.util.options")

  -- toggle options
  vim.keymap.set("n", "<leader>sn", options.toggle_number,         { desc = ui.Toggle.." toggle 'number'" })
  vim.keymap.set("n", "<leader>sr", options.toggle_relativenumber, { desc = ui.Toggle.." toggle 'relativenumber'" })
  vim.keymap.set("n", "<leader>sl", options.toggle_list,           { desc = ui.Toggle.." toggle 'list'" })
  vim.keymap.set("n", "<leader>sw", options.toggle_wrap,           { desc = ui.Toggle.." toggle 'wrap'" })
  vim.keymap.set("n", "<leader>ss", options.toggle_spell,          { desc = ui.Toggle.." toggle 'spell'" })
end
-- stylua: ignore end

return M
