local make_keymap = function(cmp)
  local if_visible = function(yes, no)
    no = no or function(fallback) fallback() end
    return function(fallback)
      if cmp.visible() then
        yes(fallback)
      else
        no(fallback)
      end
    end
  end

  local m = cmp.mapping
  local select = { behavior = cmp.SelectBehavior.Select }
  local next_or_complete = if_visible(m.select_next_item(select), m.complete())
  local prev_or_complete = if_visible(m.select_prev_item(select), m.complete())
  local next = if_visible(m.select_next_item(select))
  local prev = if_visible(m.select_prev_item(select))

  local abort = if_visible(m.abort())
  local confirm = if_visible(m.confirm { select = true })
  local confirm_or_complete = if_visible(m.confirm { select = true }, m.complete())

  local scroll_docs_down = m.scroll_docs(3)
  local scroll_docs_up = m.scroll_docs(-3)

  -- Mappings that should work in both command line and Insert mode.
  return {
    -- stylua: ignore start
    ["<c-n>"]    = { i = next_or_complete, c = next_or_complete },
    ["<c-p>"]    = { i = prev_or_complete, c = prev_or_complete },

    ["<c-j>"]    = { i = next_or_complete, c = next },
    ["<c-k>"]    = { i = prev_or_complete, c = prev },

    ["<down>"]   = { i = next_or_complete, c = next },
    ["<up>"]     = { i = prev_or_complete, c = prev },

    ["<s-down>"] = { i = scroll_docs_down, c = scroll_docs_down },
    ["<s-up>"]   = { i = scroll_docs_up,   c = scroll_docs_up },

    ["<c-c>"]    = { i = abort,   c = abort },
    ["<c-e>"]    = { i = abort,   c = abort },
    ["<c-y>"]    = { i = confirm, c = confirm },

    ["<tab>"]    = { i = confirm, c = confirm_or_complete },
    -- stylua: ignore end
  }
end

return {
  "hrsh7th/nvim-cmp",

  dependencies = {
    "hrsh7th/cmp-nvim-lsp",
    "hrsh7th/cmp-nvim-lua",
    "hrsh7th/cmp-path",
    "hrsh7th/cmp-buffer",
    "hrsh7th/cmp-cmdline",
    "onsails/lspkind-nvim",

    "L3MON4D3/LuaSnip",
    "saadparwaiz1/cmp_luasnip",
  },

  event = { "CmdlineEnter", "InsertEnter" },

  config = function()
    local cmp = require("cmp")
    local keymap = make_keymap(cmp)

    cmp.setup {
      mapping = keymap,

      enabled = function()
        local ctx = require("cmp.config.context")
        return not ctx.in_treesitter_capture("comment")
          and not ctx.in_syntax_group("Comment")
      end,

      snippet = {
        expand = function(args) require("luasnip").lsp_expand(args.body) end,
      },

      formatting = {
        format = require("lspkind").cmp_format {
          mode = "symbol_text",
          symbol_map = require("fschauen.util.icons").kind,
          menu = {
            buffer = "buf",
            nvim_lsp = "LSP",
            nvim_lua = "lua",
            path = "",
          },
        },
      },

      sources = cmp.config.sources({
        { name = "nvim_lua" },
        { name = "nvim_lsp" },
        { name = "luasnip" },
      }, {
        { name = "path" },
        { name = "buffer", keyword_length = 5 },
      }),

      window = {
        completion = cmp.config.window.bordered(),
        documentation = cmp.config.window.bordered(),
      },

      experimental = { ghost_text = true },
    }

    cmp.setup.cmdline(":", {
      mapping = keymap,
      completion = { autocomplete = false },
      sources = cmp.config.sources({
        { name = "path" },
      }, {
        { name = "cmdline" },
      }),
    })

    cmp.setup.filetype("TelescopePrompt", {
      enabled = false,
    })
  end,
}
